/*
 *	     Module : SA9903.c
 *              pjt : SA9903.pjt
 *      description :   PIC based code to implement a 9903 based energy-consumption meter.
 * 			 Uses 'native' PIC SPI @ a 312.5 kHz clock. 
 *                     
 *              ver : 1.0
 *	         log:
 *		1.0 _df 29-09-2002 init coding
 */

#define CRYSTAL_DIV256	3579545./256.	//constant in frequency calculation, see application note.

//Compilation Flags


//project's includes
#include <stdio.h>
#include <stdlib.h>
#include <math.h>			//using floor() & ceil()
#include "sa9903.h"			//project's header file


/*isr()
 * 1. 10ms ticks via tmr1 ctrls the LEDs
 * 2. rising edge on RB0 (INT) raises sample bitflag 
 * 3. key i/p via RBIF, B5..B7
 */
static void interrupt isr(void)
{
  
  if(CCP1IF)			/* 10ms ticks */
  {
     if(!(aLED|rLED)) t_LED--;	//decr led timer if supposed to be off
     if(!t_LED)			//LED (off) ctrl, 10ms after switching it on switch it off
      {
       if((!oA_LED)&&(!aLED)) oA_LED=ON;
       if((!oR_LED)&&(!rLED)) oR_LED=ON;
       t_LED=PULSE; 		//rst pulsewidth counter
      }

     if(aLED) 			//LED (on) ctrl, switch on and reset bitflag
      {
       if(oA_LED) oA_LED=OFF;
       aLED=OFF;
      }
     if(rLED)
      {
       if(oR_LED) oR_LED=OFF;
       rLED=OFF;
      }

      if (tick_count == SECOND) /* 1s ticks */
       {
	display = ON;		//update LCD
        tick_count=0;		//rst tick_count
       }
      tick_count++;		//count ticks
    CCP1IF=OFF;			//rst interrupt flag
  }

  if(INTF) 			/* zero crossing */
   {
    sample=ON;			//read registers
    INTF=OFF;			//rst interrupt flag
   }

  if(RBIF) 			/* keys */
   {
    tempB=PORTB;

    if((tempB & KEY_MASK) == KEY_MASK)
     {
      if(!iENT)
       {
        key=ENTER;
	ent_menu^=ON; 	//toggle enter menu's bitflag
       }
       else if(!iUP)   key=UP;
        else if(!iDWN)
         {
          key=DOWN;
	  dwn_menu^=ON; //toggle down menu's bitflag
         }

         else 	       key=0;	//weird accident?
     }				//else we do nothing ->key keeps previous value
    RBIF=OFF;			//rst interrupt flag
   }
}

void init(void)
{
  uchar cnt;			//counter for initialise loop

  //option reg
  OPTION = 0b01000000;		//PortB pull-up, INT rising edge rest off
  T1CON  = SETUP_TMR1; 		//tmr1 setup

/*Port A set-up */
  ADCON1=0x06; 			//all digital i/o
  ADCON0=0x00; 			//not using A/D

  //setup ports
  TRISA = IO_BITS_A;
  TRISB = IO_BITS_B;
  TRISC = IO_BITS_C;

//configure 16F876's spi

//MOSI rising edge + MISO falling edge + Sclk=(Fosc/64)==312.5 kHz
  SSPSTAT = 0xc0;		//SMP=1+CKE=0
  SSPCON  = 0x22;		//CKP=0+spi mode


//soft defaults
  oLCD        = OFF;	  	//LCD's Chip Select low
  oMEM        = OFF;    	//e^2's Chip Select low
  oMTR        = OFF;	  	//metering ic's Chip Select low
  aLED	      = OFF;		//active-led ctrl bitflag
  rLED        = OFF;		//re-active-led ctrl bitflag
  oA_LED      = ON;	  	//active    rate LED output is off
  oR_LED      = ON;	  	//re-active rate LED output is off 
  t_LED       = PULSE; 		//led pulsewidth counter
  sample      = OFF;		//not ready to acquire new data
  tick_count  = 0;       	//clear 10ms tick counter
  dwn_menu    = OFF;		//voltage
  ent_menu    = OFF;		//starts on kWh
  key         = 0;		//LCD starts on kWh
  display     = ON;             //immediately display something

  for(cnt=0;cnt<2;cnt++)
   {
    accumulator[cnt] = 0;				//clr accumulators
    threshold[cnt]   = (long)( THRES*(1.+CAL[cnt]) );	//setup active/reactive thresholds
    first[cnt]       = ON; 				//is first iteration
    energy[cnt]      = 0.0;				//display starts as zero
   }

//setup interrupts
	TMR1ON  = OFF;
	TMR1H   = 0x0;		//clr tmr1
	TMR1L   = 0x0;
	CCPR1H  = LOAD_CCP1H;	//load CCP1 Hi
	CCPR1L  = LOAD_CCP1L;	//load CCP1 Lo
	CCP1CON = SETUP_CCP1;	//CCP1 setup
	CCP1IF  = OFF; 		
	CCP1IE  = ON;
	TMR1ON  = ON;
	INTE	= ON;		//ext interrupt (RB0)
  	PEIE    = ON; 		//enable peripheral interrupts
	RBIE    = ON;		//PortB interrupt on change
	GIE     = ON;		//global interrupt enabled
}

/* labs(long x)
 * returns abs of x
 */
long labs(long val)
{
  if(val<0) 
    return(-val);
  return(val);
}

/*next fcn checks for register wrapping 
 * feed fcn with (new_val-old_val),
 *  then just 'save' result in accumulator...
 */
long wrap(long value)
{
 signed char sign;
  
  test=labs(value);
  if( test > 0x7fffff)    	 		//val>8388607?
   {
     if(value>0) sign=-1;
            else sign= 1;
     value=(0x1000000+(sign*value))*sign;	//(16777216(+or-)value)*(+or-)
   }
  return (value);
}

/* process_data()
 * does active and reactive thresholding,
 *  raising the LED-to-pulse bitflag if required
 */
void process_data(void)
{
  test=labs(accumulator[ACT]);
  if((test > threshold[ACT]))
   {
    if((!aLED)&&(oA_LED))	aLED=ON; //if led not on signal that it must be

    if(accumulator[ACT] > 0)	accumulator[ACT]-= threshold[ACT];
    else  		 	accumulator[ACT]+= threshold[ACT];
    energy[ACT]+=ENG_PULSE;
   }

  test=labs(accumulator[REACT]);
  if((test > threshold[REACT]))
   {
    if((!rLED)&&(oR_LED))	rLED=ON; //if led not on signal that it must be 

    if(accumulator[REACT] > 0)	accumulator[REACT]-= threshold[REACT];
    else  		 	accumulator[REACT]+= threshold[REACT];
    energy[REACT]+=ENG_PULSE;
   }
}

/*boot_scrn()
 *on power-up displays 
 * chip-type/mtr/ver#-> run
 */
void boot_scrn()
{
 uchar	b_buff[8],	//temp lcd buffer
	cnt;		//for loop counter

  lcd_clear();		//start afresh
  for(cnt=CHIP;cnt<(RUN+1);cnt++)
   {
    sprintf(b_buff,"%s",Display[cnt]);
    lcd(b_buff,P_WIDTH );
    prompt_delay();
    lcd_clear();
   }
}

/*main()
 *initialise uCtrler/software then does boot screen 
 *data acquisition sequentially
 * local variables for acquisition and menu
 */
void main(void)
{
  uchar	address,		//register to be read
	old_key, select;	//used in if(display)

  init();
  boot_scrn();

  old_key=ENTER; // the start-up selection

  while(1)
  {
    if(sample) //read all registers starting from active 
     {
       address=ACT;				//start address
	 new_val= acquire(address);		//active    register
	 if(!first[ACT])			//if this is not the 1st sample
	 {
	  delta_val= new_val - prev_val[ACT];	//delta_val=new_val-previous_val
    	  accumulator[ACT]+= wrap(delta_val);	//add to accumulator
	 }
	 else first[ACT] = OFF;			//we've completed 1st sample
	 prev_val[ACT] = new_val;		//store new_val


	 new_val= acquire(++address);		//reactive  register
	 if(!first[REACT])			//if this is not the 1st sample
	 {
	  delta_val= new_val - prev_val[REACT];	//delta_val=new_val-previous_val
    	  accumulator[REACT]+= wrap(delta_val);	//add to accumulator
	 }
	 else first[REACT] = OFF;		//we've completed 1st sample
	 prev_val[REACT] = new_val;		//store new_val

	 volt= acquire(++address);		//voltage   register


	 freq= (acquire(++address) & 0x0003ff); //frequency register,only intersted in last 10 bits 

	 sample= OFF;				//rst flag
         process_data();			//threshold & LEDs
     }


    if(display)
     {
      if((key)&&(key!=old_key)) select=key;
       else select=old_key;

      switch(select)
       {	  
/*
        case UP:

 reserved for calibration sub-menu

*/
	case DOWN:		//Volt/Freq display
		   if(dwn_menu) 		//frequency 
		    {
		     lcd_temp=floor( ( CRYSTAL_DIV256 ) / ( (float) freq ) );
		     num_char= sprintf(disp_buff,"%3.1f",lcd_temp);
		     lcd(disp_buff,num_char);
		     lcd_icon(HZ);		//show icon
		    }
		   else				//voltage 
		    {
		     lcd_temp=(float) volt;
		     lcd_temp /=780.;		//2B set via cal/setup, _df: hand cal
                     lcd_temp *=230.;		

		     num_char= sprintf(disp_buff,"%5.1f",lcd_temp);
		     lcd(disp_buff,num_char);
		     lcd_icon(VOLT);		//show icon
		    }
		   break;

	case ENTER:		//kWh/kVarh display
		   if(ent_menu) 		//kVarh
		    {
		     num_char= sprintf(disp_buff,"%1.3f",energy[REACT]);
		     lcd(disp_buff,num_char);
		     lcd_icon(KVAR);		//show icon
		    }
		   else				//kWh
		    {
		     num_char= sprintf(disp_buff,"%1.3f",energy[ACT]); 
		     lcd(disp_buff,num_char);
		     lcd_icon(KWH);		//show icon
		    }
		   break;
       }
	old_key=select;
       display=OFF; //rst flag
     }
  }
}
